/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: IIOPAddress.java,v 1.7 2001/08/13 03:47:33 wood Exp $
 *
 * Date         Author  Changes
 */

package org.openorb.iiop;

import java.util.ArrayList;

import org.openorb.net.AbstractAddress;

import java.net.*;

import org.omg.IOP.*;
import org.omg.IIOP.*;

import org.omg.CORBA.CompletionStatus;

import org.omg.CONV_FRAME.CodeSetComponent;
import org.omg.CONV_FRAME.CodeSetComponentInfo;
import org.omg.CONV_FRAME.CodeSetComponentInfoHelper;
import org.openorb.net.Address;

/** 
 * IIOP request target. 
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.7 $ $Date: 2001/08/13 03:47:33 $ 
 */
public class IIOPAddress extends AbstractAddress implements Cloneable {

  private IIOPAddress() {
  }

	public void setHost(InetAddress host) {
		if(host != null) {
			_host = host;
			_hostname = host.getHostName();
			_hash = -1;
		}
	}
	
	public void setPort(int port) {
		if(port != 0) {
			_port = port;
			_hash = -1;
		}
	}

  public void setProtocol(String protocol) {
    _protocol = protocol;
    _hash = -1;
  }
  
	public Object clone() {
		try { 
			return super.clone();
		} 
		catch(CloneNotSupportedException ex) { 
			org.openorb.util.Trace.assertTrue(false);
			return null;
		}
	}
	
  private InetAddress _host;
  private String _hostname;
  private int _port;
  private Version _vers;  
  private int _orb_type = 0;
	private int _hash = -1;
  private String _protocol = "iiop";
	
	private CodeSetComponentInfo _codesetComponentInfo = null;
	
  private IIOPAddress [] alternates = null;

	/**
	 * Mask of bits set for the version priority.
	 */
	public static final short MASK_PRI_VERSION = 0x000F;
	
	/**
	 * Version priorities
	 */
	public static final short [] PRI_VERSION = { 0x000F, 0x000E, 0x000D };
	
	/**
	 * Flag set if address is an alternate address.
	 */
	public static final short PRI_FLAG_ALTERNATE = 0x0800;
	
	/**
	 * Flag set if address is from profile not containing primary profile 
	 * component.
	 */
	public static final short PRI_FLAG_NONPRIMARY = 0x0400;
	
	/**
	 * Bits available to be altered for differing security levels.
	 */
	public static final short MASK_PRI_SECURE = 0x00F0;
		
	/**
	 * Standard unsecured address.
	 */
	public static final short PRI_UNSECURE = 0x00F0;
	
  /**
   * The protocol string as would appear in the corbaloc address.
   */
  public String getProtocol() {
    return _protocol;
  }
  
	/**
	 * Return the hostname of the target.
	 */
  public String get_hostname() {
    return _hostname;
  }
  
	/**
	 * Return the host as an InetAddress
	 */
  public InetAddress get_host() {
    return _host;
  }
  
	/**
	 * Return the port number of the target.
	 */
  public int get_port() {
    return _port;
  }
  
	/**
	 * Get the IIOP version of the target.
	 */
  public Version get_version() {
    return _vers;
  }
  
	/**
	 * Return the four char orb type.
	 */
  public int get_orb_type() {
    return _orb_type;
  }
  
	/**
	 * Return the codeset data.
	 */
	public CodeSetComponentInfo getCodesetComponentInfo() {
		return _codesetComponentInfo;
	}
	
  /**
   * Return all the alternative addresses created from this profile.
   */
  public Address [] get_alternates() {
    return alternates;
  }
  
  /**
   * A string which could be used to contact the endpoint in a corbaloc 
   * style address.
   */
  public String getEndpointString() {
    return _protocol + ":" + _vers.major + "." + _vers.minor + "@" + _hostname + ":" + _port;
  }
  
  /**
   * Human readable string describing the endpoint.
   */
  public String getEndpointDescription() {
    StringBuffer sb = new StringBuffer();
    sb.append(getEndpointString());
    if(_host == null)
      sb.append(" host unknown");
    else {
      boolean comma = false;
      if(getPriority(MASK_PRI_SECURE) != PRI_UNSECURE) {
        sb.append(" flags: secure(");
        sb.append(getPriority(MASK_PRI_SECURE) >> 8);
        sb.append(")");
        comma = true;
      }
      if(getPriority(PRI_FLAG_NONPRIMARY) == 0) {
        if(comma)
          sb.append(", ");
        else
          sb.append(" flags: ");
        sb.append("primary");
        comma = true;
      }
      if(getPriority(PRI_FLAG_ALTERNATE) != 0) {
        if(comma)
          sb.append(", ");
        else
          sb.append(" flags: ");
        sb.append("alternate");
      }
    }
    return sb.toString();
  }
      
	public int hashCode() {
		if(_hash == -1)
			_hash = super.hashCode() ^ _host.hashCode() ^ _port ^ _protocol.hashCode();
		return _hash;
	}
	
	public boolean equals(Object obj) {
		if(!(obj instanceof IIOPAddress))
			return false;
	
		IIOPAddress ia2 = (IIOPAddress)obj;
    
    if(_port != ia2._port || !_hostname.equals(ia2._hostname) || !_protocol.equals(ia2._protocol) ||
       _vers.major != ia2._vers.major || _vers.minor != ia2._vers.minor)
      return false;
    
    return super.equals(obj);
	}
	
	/**
	 * Create a set of IIOP addresses from a profile in an IOR.
	 */
  static IIOPAddress [] get_addresses(org.omg.IOP.Codec codec, IOR ior, int selected_profile) 
	{
    if(ior.profiles[selected_profile].tag != TAG_INTERNET_IOP.value)
      throw new org.omg.CORBA.INV_OBJREF("Invalid tag for object reference", 
																					IIOPMinorCodes.INV_OBJREF_BAD_TAG, CompletionStatus.COMPLETED_NO);
  
    IIOPAddress addr = new IIOPAddress();
    addr.set_ior(ior, selected_profile);
        
    byte [] pd = ior.profiles[selected_profile].profile_data;
    
   
    if(pd[1] != 1)
      return new IIOPAddress[0];

    TaggedComponent [] procomp = null;
    
    try {			
      switch(pd[2]) {
      case 0: {
        // IIOP 1.0
        ProfileBody_1_0 body = ProfileBody_1_0Helper.extract(codec.decode_value(pd, ProfileBody_1_0Helper.type()));
        addr._vers = body.iiop_version;
        addr._hostname = body.host;
        addr._port = (body.port&0xFFFF);
	/*
	  for(int i=0;i<body.object_key.length;i++)
	  System.out.print(body.object_key[i]+" ");
	  System.out.println("##***");
	*/
	addr.set_oid(body.object_key);
				procomp = new TaggedComponent[0];
      } break;
      case 1:
      case 2:
      default: {
        ProfileBody_1_1 body = ProfileBody_1_1Helper.extract(codec.decode_value(pd, ProfileBody_1_1Helper.type()));
        addr._vers = body.iiop_version;
        addr._hostname = body.host;
        addr._port = (body.port&0xFFFF);
	/*
	  for(int i=0;i<body.object_key.length;i++)
	  System.out.print(body.object_key[i]+" ");
	  System.out.println("##***");
	*/
        addr.set_oid(body.object_key);
        procomp = body.components;
				
				if(addr._vers.minor > 2)
					addr._vers.minor = 2;
      } break;
      }
      try {
        addr._host = InetAddress.getByName(addr._hostname);
      }
      catch(UnknownHostException ex) {
        addr._host = null;
				// set hash since trying to set it will fail.
				addr._hash = addr._hostname.hashCode() ^ addr._port;
      }
      
      // scan for multiple components
      int mcc = 0;
      for(int i = 0; i < ior.profiles.length; ++i)
        if(ior.profiles[i].tag == TAG_MULTIPLE_COMPONENTS.value)
          ++mcc;

      TaggedComponent [] cpts;
      if(mcc == 0)
        cpts = procomp;
      else {
        TaggedComponent [][] mcs;
        if(procomp != null) {
          mcs = new TaggedComponent[mcc+1][];
          mcs[mcc] = procomp;
        } else {
          mcs = new TaggedComponent[mcc][];
        }

        for(int i = 0; i < ior.profiles.length && mcc > 0; ++i)
          if(ior.profiles[i].tag == TAG_MULTIPLE_COMPONENTS.value)
            mcs[--mcc] = TaggedComponentSeqHelper.extract(codec.decode_value(ior.profiles[i].profile_data, TaggedComponentSeqHelper.type()));

        mcc = 0;
        for(int i = 0; i < mcs.length; ++i)
          mcc += mcs[i].length;

        cpts = new TaggedComponent[mcc];

				mcc = 0;
				for(int i = 0; i < mcs.length; ++i) {
					System.arraycopy(mcs[i], 0, cpts, mcc, mcs[i].length);
					mcc += mcs[i].length;
				}
      }

      addr.set_components(cpts, procomp.length);

      org.omg.CORBA.ORB sorb = org.omg.CORBA.ORB.init();
      			
			short priority = (short)(PRI_FLAG_NONPRIMARY | PRI_UNSECURE | PRI_VERSION[pd[2]]);
			
      ArrayList alts = new ArrayList();
      for(int i = 0; i < cpts.length; ++i) {
        switch(cpts[i].tag) {
        case TAG_ALTERNATE_IIOP_ADDRESS.value: {
          ListenPoint lp = ListenPointHelper.extract(codec.decode_value(cpts[i].component_data, ListenPointHelper.type()));
          alts.add(lp);
          addr.set_component_data(i, "ComponentID = " + cpts[i].tag + " Alternate IIOP Address: " + lp.host + ":" + lp.port);
        } break;
        case TAG_ORB_TYPE.value:
          addr._orb_type = codec.decode_value(cpts[i].component_data, sorb.get_primitive_tc(org.omg.CORBA.TCKind.tk_long)).extract_long();
          String val = Integer.toHexString(addr._orb_type);
          addr.set_component_data(i, "ComponentID = " + cpts[i].tag + " ORB Type: 0x" + "00000000".substring(val.length()) + val);
          break;
        case org.omg.IOP.TAG_CODE_SETS.value: {
					addr._codesetComponentInfo = CodeSetComponentInfoHelper.extract(codec.decode_value(cpts[i].component_data, CodeSetComponentInfoHelper.type()));
          StringBuffer sb = new StringBuffer("ComponentID = " + cpts[i].tag + " Codesets \n");
          sb.append("\tFor char Data: ");
          int cs = addr._codesetComponentInfo.ForCharData.native_code_set;
          for(int j = 0; j <= addr._codesetComponentInfo.ForCharData.conversion_code_sets.length; ++j) {
            String css = Integer.toHexString(cs);
            sb.append("\n\t\t0x").append("00000000".substring(css.length())).append(css);
            sb.append(" ").append(CodeSetDatabase.codesetIDtoDescription(cs));
            if(j < addr._codesetComponentInfo.ForCharData.conversion_code_sets.length)
              cs = addr._codesetComponentInfo.ForCharData.conversion_code_sets[j];
          }
          sb.append("\n\tFor wchar Data: ");
          cs = addr._codesetComponentInfo.ForWcharData.native_code_set;
          for(int j = 0; j <= addr._codesetComponentInfo.ForWcharData.conversion_code_sets.length; ++j) {
            String css = Integer.toHexString(cs);
            sb.append("\n\t\t0x").append("00000000".substring(css.length())).append(css);
            sb.append(" ").append(CodeSetDatabase.codesetIDtoDescription(cs));
            if(j < addr._codesetComponentInfo.ForWcharData.conversion_code_sets.length)
              cs = addr._codesetComponentInfo.ForWcharData.conversion_code_sets[j];
          }
          addr.set_component_data(i, sb.toString());
        } break;
				case 28: /* org.omg.FT.TAG_FT_PRIMARY */
					// this just contains a boolean.
          if(cpts[i].component_data[1] != 0) {
						priority = (short)(priority & ~PRI_FLAG_NONPRIMARY);
            addr.set_component_data(i, "ComponentID = " + cpts[i].tag + " Primary profile");
          } else {
            addr.set_component_data(i, "ComponentID = " + cpts[i].tag + " NonPrimary profile");
          }
					break;
        case TAG_POLICIES.value:
          // TODO:
          throw new org.omg.CORBA.NO_IMPLEMENT();
        }
      }

			addr.setPriority(priority);

			// if anything else is done do it before doing this
      IIOPAddress [] addrs = new IIOPAddress[1+alts.size()];
      addrs[0] = addr;
      for(int i = 0; i < alts.size(); ++i) {
        ListenPoint lp = (ListenPoint)alts.get(i);
				addrs[i+1] = (IIOPAddress)addr.clone();
        addrs[i+1]._hostname = lp.host;
        try {
          addrs[i+1]._host = InetAddress.getByName(lp.host);
        }
        catch(UnknownHostException ex) {
          addrs[i+1]._host = null;
					// set hash since trying to set it will fail.
					addrs[i+1]._hash = addrs[i+1]._hostname.hashCode() ^ addrs[i+1]._port;
        }
        addrs[i+1]._port = lp.port&0xFFFF;
				addrs[i+1].setPriority(PRI_FLAG_ALTERNATE, PRI_FLAG_ALTERNATE);
      }

      return addrs;
    }
    catch(org.omg.IOP.CodecPackage.FormatMismatch ex) {
      throw new org.omg.CORBA.INV_OBJREF("Profile data cannot be parsed", 
																					IIOPMinorCodes.INV_OBJREF_BAD_PROFILE, CompletionStatus.COMPLETED_NO);
    }
    catch(org.omg.IOP.CodecPackage.TypeMismatch ex) {
      throw new org.omg.CORBA.INV_OBJREF("Profile data cannot be parsed", 
																					IIOPMinorCodes.INV_OBJREF_BAD_PROFILE, CompletionStatus.COMPLETED_NO);
    }    
  } 
}
