/**
* Redistribution and use of this software and associated
* documentation ("Software"), with or without modification, are
* permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain copyright statements
*    and notices.  Redistributions must also contain a copy of this
*    document.
*
* 2. Redistributions in binary form must reproduce the above
*    copyright notice, this list of conditions and the following
*    disclaimer in the documentation and/or other materials provided
*    with the distribution.
*
* 3. The name "Exolab" must not be used to endorse or promote
*    products derived from this Software without prior written
*    permission of Intalio Inc.  For written permission, please
*    contact info@exolab.org.
*
* 4. Products derived from this Software may not be called "Exolab"
*    nor may "Exolab" appear in their names without prior written
*    permission of Intalio Inc. Exolab is a registered trademark of
*    Intalio Inc.
*
* 5. Due credit should be given to the Exolab Project
*    (http://www.exolab.org/).
*
* THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
* THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
* PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
* ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
* OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*
* Copyright 2000 (C) Intalio Inc. All Rights Reserved.
*
* $Id: ClientBinding.java,v 1.4 2001/07/31 13:45:34 wood Exp $
*
* Date         Author  Changes
*/

package org.openorb.net;

import org.openorb.adapter.*;

import java.util.Comparator;
import org.omg.CORBA.CompletionStatus;

/** 
* A client binding represents a potential request path between client
* and server.
*
* @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
* @version $Revision: 1.4 $ $Date: 2001/07/31 13:45:34 $ 
*/
public class ClientBinding {
    protected ClientBinding(Address address) {
	_address = address;
	_forcedMarshal = FORCE_UNSET;

	_priority = (int)address.getPriority();
	_version=0;
    }

    /**
     * Bind address to remote invocation channel.
     */
    public ClientBinding(Address address, ClientChannel channel, 
			 Object transportAssociation) {
	_address = address;
	_channel = channel;
	_transportAssociation = transportAssociation;

	_forcedMarshal = FORCE_TRUE;

	_priority = (int)address.getPriority();
	_version=0;
    }

    /**
     * Bind address to local adapter.
     */
    public ClientBinding(Address address, ClientChannel channel, 
			 Object transportAssociation, ServerManager serverManager) {
	_address = address;
	_channel = channel;
	_transportAssociation = transportAssociation;
	_serverManager = serverManager;

	_forcedMarshal = FORCE_UNSET;

	_priority = (int)address.getPriority();
	_version=0;
    }

    /**
     * Bind address to a system exception. Placeholder for failed bindings
     */
    public ClientBinding(Address address, org.omg.CORBA.SystemException ex) {
	_address = address;
	_exception = ex;
	_priority = DEAD_PRIORITY;
	_forcedMarshal = FORCE_FALSE;
	_version=0;
    }


    protected ClientBinding(Address address,int version) {
	_address = address;
	_forcedMarshal = FORCE_UNSET;

	_priority = (int)address.getPriority();
	_version=version;
    }

    /**
     * Bind address to remote invocation channel.
     */
    public ClientBinding(Address address, ClientChannel channel, 
			 Object transportAssociation,int version) {
	_address = address;
	_channel = channel;
	_transportAssociation = transportAssociation;

	_forcedMarshal = FORCE_TRUE;

	_priority = (int)address.getPriority();
	_version=version;
    }

    /**
     * Bind address to local adapter.
     */
    public ClientBinding(Address address, ClientChannel channel, 
			 Object transportAssociation, ServerManager serverManager,int version) {
	_address = address;
	_channel = channel;
	_transportAssociation = transportAssociation;
	_serverManager = serverManager;

	_forcedMarshal = FORCE_UNSET;

	_priority = (int)address.getPriority();
	_version=version;
    }

    /**
     * Bind address to a system exception. Placeholder for failed bindings
     */
    public ClientBinding(Address address, org.omg.CORBA.SystemException ex,int version) {
	_address = address;
	_exception = ex;
	_priority = DEAD_PRIORITY;
	_forcedMarshal = FORCE_FALSE;
	_version=version;
    }


    public static final int DEAD_PRIORITY = 0x7FFFFFFF;
    private Address _address;
    private ClientChannel _channel;
    private Object _transportAssociation;
    private ObjectAdapter _adapter = null;
    private org.omg.CORBA.SystemException _exception;
    private int _version;  /* Field added by Diana on August 23*/
    private int _priority = 0;
    private byte [] _oid = null;
    private int _forcedMarshal;
    private ServerManager _serverManager;
    private static final int FORCE_UNSET = -1;
    private static final int FORCE_FALSE = 0;
    private static final int FORCE_TRUE = 1;
    /**
     * Return the client address.
     */
    public Address getAddress() {
	return _address;
    }

    /**
     * Return the client channel. This is always available for both local and
     * nonlocal bindings.
     */
    public ClientChannel getClientChannel() {
	return _channel;
    }

    /**
     * Return the object adapter. This will return null if the binding is non-local.
     */
    public ObjectAdapter getObjectAdapter() {
	if(_serverManager == null)
	    return null;
	if(_oid == null)
	    _oid = _address.getTargetAddress(org.omg.GIOP.KeyAddr.value).object_key();
	_adapter = _serverManager.find_adapter(_oid);
	if(_adapter == null)
	    return null;
	_forcedMarshal = FORCE_UNSET;
	return _adapter;
    }

    /**
     * Get binding priority. Low priorities are used first.
     * The lowest order byte of the priority is inter-component priority,
     * The second order byte is inter-profile priority,
     * The third order byte is inter-refrerence prioriy
     * The highest order byte is the channel state / adapter priority.
     */
    public int getPriority() {
	if(_exception != null)
	    return DEAD_PRIORITY;
	if (_adapter != null)
	    return _priority | (_adapter.cache_priority() & 0xFF000000);
	else
	    return _priority | _channel.state();
    }  

    /**
     * Set the binding priority. Lower values are used first.
     * This should be set with respect to the current priority, changing only
     * the appropriate bits.
     */
    public void setPriority(int priority) {
	if(priority > 0)
	    _priority = priority & (MASK_PROFILE_PRIORITY | MASK_IOR_PRIORITY) | 
		_priority & ~(MASK_PROFILE_PRIORITY | MASK_IOR_PRIORITY);
	else
	    _priority = priority;
    }

    /**
     * Significant bits in selecting between alternate addresses from the
     * same profile. These bits are not settable by the setPriority operation.
     */
    public static final int MASK_ADDRESS_PRIORITY = 0x0FFF;

    /**
     * Significant bits in selecting between alternate addresses from different
     * profiles.
     */
    public static final int MASK_PROFILE_PRIORITY = 0xF000;

    /**
     * Significant bits in selecting between alternate IORs
     */
    public static final int MASK_IOR_PRIORITY = 0xFF0000;  

    public int hashCode() {
	return _address.hashCode();
    }

    public boolean equals(Object obj) {
	if(!(obj instanceof ClientBinding))
	    return false;

	return _version==((ClientBinding)obj)._version && _address.equals(((ClientBinding)obj)._address);
    }

    /**
     * True if the target can accept local invocations. The object reference
     * operations such as is_a are available for local invocation whenever
     * getObjectAdapter returns non null.
     */
    public boolean local_invoke() {
	if(_forcedMarshal == FORCE_UNSET) {
	    while(true) {
		getObjectAdapter();
		if(_adapter == null) {
		    _forcedMarshal = FORCE_TRUE;
		    break;
		}
		try {
		    _forcedMarshal = _adapter.forced_marshal(_oid)?FORCE_TRUE:FORCE_FALSE;
		    break;
		}
		catch(org.openorb.adapter.AdapterDestroyedException ex) {
		    _adapter = null;
		}
	    }
	}
	return _forcedMarshal == FORCE_FALSE;
    }

    // request creation.

    /**
     * Create a request. If this is the first request on this channel
     * then client_connect will be called on all ChannelInterceptor
     * before returning the request. This may return null if the channel
     * is unable to create a request. This just calls create_request
     * on the channel with the binding's address and the specified target.
     *
     * @param target The target of the request.
     * @param address The target address. If the target has been redirected
     *            this may not correspond to the target's ior.
     */
    public ClientRequest create_request(org.omg.CORBA.Object target,
					String operation, 
					boolean response_expected) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_request(target, _address, _transportAssociation, operation, response_expected);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }

      /**
     * Create a request. If this is the first request on this channel
     * then client_connect will be called on all ChannelInterceptor
     * before returning the request. This may return null if the channel
     * is unable to create a request. This just calls create_request
     * on the channel with the binding's address and the specified target.
     *
     * @param target The target of the request.
     * @param address The target address. If the target has been redirected
     *            this may not correspond to the target's ior.
     */
    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,String operation, boolean response_expected) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_request(target, retention_id, client_id,_address, _transportAssociation, operation, response_expected);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }


    /* Added by Diana on feb. 2 2002*/

        public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.IOP.ServiceContext one_sc,String operation, boolean response_expected) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_request(target, retention_id, client_id,one_sc,_address, _transportAssociation, operation, response_expected);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }

    public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long request_start_time,org.omg.IOP.ServiceContext one_sc,String operation, boolean response_expected) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_request(target, retention_id, client_id,request_start_time,one_sc,_address, _transportAssociation, operation, response_expected);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }

    /* Modif by Diana on Oct. 28 2001*/

        public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long request_start_time,String operation, boolean response_expected) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_request(target, retention_id, client_id,request_start_time,_address, _transportAssociation, operation, response_expected);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }


       public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.Dynamic.Parameter[] arguments,String operation, boolean response_expected) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_request(target, retention_id, client_id,arguments,_address, _transportAssociation, operation, response_expected);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }
    /* new version on oct. 10 2001 by Diana*/

     public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,org.omg.Dynamic.Parameter[] arguments,org.omg.CORBA.TypeCode[] exceptions,org.omg.CORBA.Any result,String operation, boolean response_expected) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_request(target, retention_id, client_id,arguments,exceptions,result,_address, _transportAssociation, operation, response_expected);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }
    /* New version on Oct. 28 2001*/

     public ClientRequest create_request(org.omg.CORBA.Object target,int retention_id,String client_id,long request_start_time,org.omg.Dynamic.Parameter[] arguments,org.omg.CORBA.TypeCode[] exceptions,org.omg.CORBA.Any result,String operation, boolean response_expected) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_request(target, retention_id, client_id,request_start_time,arguments,exceptions,result,_address, _transportAssociation, operation, response_expected);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }
    /**
     * Create a locate request. This may return null if the channel
     * is unable to create a request. This just calls create_locate_request
     * on the channel with the binding's address and the specified target.
     *
     * @param target The target of the request.
     */
    public ClientRequest create_locate_request(org.omg.CORBA.Object target) 
    {
	if(_exception != null)
	    throw _exception;
	while(true) {
	    if(_channel == null)
		return null;
	    try {
		return _channel.create_locate_request(target, _address, _transportAssociation);
	    }
	    catch(RebindChannelException ex) {
		_channel = ex.getClientChannel();
		_transportAssociation = ex.getTransportAssociation();
	    }
	}
    }

    // local dispatch

    /**
     * Calls servant_preinvoke on the adapter. This may return null if a retry is
     * indicated.
     */
    public org.omg.CORBA.portable.ServantObject servant_preinvoke(String operation, Class expectedType)
	throws org.omg.PortableInterceptor.ForwardRequest 
    {
	if(_exception != null)
	    throw _exception;
	if(_serverManager == null)
	    throw new org.omg.CORBA.INTERNAL();
	ObjectAdapter adapter;
	org.omg.CORBA.portable.ServantObject servObj;
	while(true) {
	    adapter = getObjectAdapter();
	    if(adapter == null)
		return null;
	    try {
		servObj = adapter.servant_preinvoke(_oid, operation, expectedType);
		break;
	    }
	    catch(org.openorb.adapter.AdapterDestroyedException ex) {
		adapter = null;
	    }
	}
	if(servObj == null) {
	    _forcedMarshal = FORCE_UNSET;
	    return null;
	}
	LocalServant ret = new LocalServant();
	ret.adapter = adapter;
	ret.orig_serv = servObj;
	ret.servant = servObj.servant;
	return ret;
    }

    private static class LocalServant extends org.omg.CORBA.portable.ServantObject {
	ObjectAdapter adapter;
	org.omg.CORBA.portable.ServantObject orig_serv;
    }

    /**
     * Calls servant_postinvoke on the adapter.
     */
    public void servant_postinvoke(org.omg.CORBA.portable.ServantObject srvObject) {
	if(_exception != null)
	    throw _exception;
	if(!(srvObject instanceof LocalServant))
	    throw new org.omg.CORBA.INTERNAL();
	LocalServant sobj = (LocalServant)srvObject;
	sobj.adapter.servant_postinvoke(_oid, sobj.orig_serv);
    }

    /**
     * Calls locate on the adapter.
     */
    public boolean locate()
	throws org.omg.PortableInterceptor.ForwardRequest 
    {
	if(_exception != null)
	    throw _exception;
	if(_serverManager == null)
	    throw new org.omg.CORBA.INTERNAL();
	ObjectAdapter adapter;
	while(true) {
	    adapter = getObjectAdapter();
	    if(adapter == null)
		return false;
	    try {
		return adapter.locate(_oid);
	    }
	    catch(org.openorb.adapter.AdapterDestroyedException ex) {
		adapter = null;
	    }
	}
    }

    /**
     * Calls is_a on the adapter.
     */
    public boolean is_a(String repository_id)
	throws org.omg.PortableInterceptor.ForwardRequest 
    {
	if(_exception != null)
	    throw _exception;
	if(_serverManager == null)
	    throw new org.omg.CORBA.INTERNAL();
	ObjectAdapter adapter;
	while(true) {
	    adapter = getObjectAdapter();
	    if(adapter == null)
		throw new org.omg.CORBA.OBJECT_NOT_EXIST(0, CompletionStatus.COMPLETED_NO);
	    try {
		return adapter.is_a(_oid, repository_id);
	    }
	    catch(org.openorb.adapter.AdapterDestroyedException ex) {
		adapter = null;
	    }
	}
    }

    /**
     * Calls get_interface_def on the adapter.
     */
    public org.omg.CORBA.Object get_interface_def()
	throws org.omg.PortableInterceptor.ForwardRequest 
    {
	if(_exception != null)
	    throw _exception;
	if(_serverManager == null)
	    throw new org.omg.CORBA.INTERNAL();
	ObjectAdapter adapter;
	while(true) {
	    adapter = getObjectAdapter();
	    if(adapter == null)
		throw new org.omg.CORBA.OBJECT_NOT_EXIST(0, CompletionStatus.COMPLETED_NO);
	    try {
		return adapter.get_interface_def(_oid);
	    }
	    catch(org.openorb.adapter.AdapterDestroyedException ex) {
		adapter = null;
	    }
	}
    }

    /**
     * Calls get_domain_managers on the adapter.
     */
    public org.omg.CORBA.DomainManager [] get_domain_managers()
	throws org.omg.PortableInterceptor.ForwardRequest 
    {
	if(_exception != null)
	    throw _exception;
	if(_serverManager == null)
	    throw new org.omg.CORBA.INTERNAL();
	ObjectAdapter adapter;
	while(true) {
	    adapter = getObjectAdapter();
	    if(adapter == null)
		throw new org.omg.CORBA.OBJECT_NOT_EXIST(0, CompletionStatus.COMPLETED_NO);
	    try {
		return adapter.get_domain_managers(_oid);
	    }
	    catch(org.openorb.adapter.AdapterDestroyedException ex) {
		adapter = null;
	    }
	}
    }

    /**
     * Calls get_component on the adapter.
     */  
    public org.omg.CORBA.Object get_component()
	throws org.omg.PortableInterceptor.ForwardRequest
    {
	if(_exception != null)
	    throw _exception;
	if(_serverManager == null)
	    throw new org.omg.CORBA.INTERNAL();
	ObjectAdapter adapter;
	while(true) {
	    adapter = getObjectAdapter();
	    if(adapter == null)
		throw new org.omg.CORBA.OBJECT_NOT_EXIST(0, CompletionStatus.COMPLETED_NO);
	    try {
		return adapter.get_component(_oid);
	    }
	    catch(org.openorb.adapter.AdapterDestroyedException ex) {
		adapter = null;
	    }
	}
    }

    private static class PriorityComparator 
	implements Comparator 
    {
	public int compare(final Object obj1, final Object obj2) {
	    if(obj1==null) {
		if(obj2==null)
		    return 0;
		return 1;
	    }
	    if(obj2 == null)
		return -1;

	    int pri1 = ((ClientBinding)obj1).getPriority();
	    int pri2 = ((ClientBinding)obj2).getPriority();

	    if(pri1 < 0) {
		if(pri2 < 0)
		    return 0;
		return -1;
	    }
	    if(pri2 < 0)
		return 1;

	    int diff = pri2 - pri1;
	    return (pri1>pri2)?1:((pri1<pri2)?-1:0);
	}

	public boolean equals(final java.lang.Object p1) {
	    return (p1 == this);
	}
    }

    final public static Comparator priorityComp = new PriorityComparator();
}
