/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio Inc.  For written permission, please
 *    contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio Inc. Exolab is a registered trademark of
 *    Intalio Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL INTALIO OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: ServerManagerImpl.java,v 1.8 2001/08/13 03:47:34 wood Exp $
 *
 * Date         Author  Changes
 */

package org.openorb.net;

import org.openorb.adapter.*;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;

import java.util.LinkedList;
import java.util.ArrayList;

import java.util.Iterator;

import org.omg.PortableServer.POAManagerPackage.State;

import org.openorb.CORBA.MinorCodes;
import org.omg.CORBA.CompletionStatus;
import org.openorb.util.MergeStack;

/**
 * This class is the controler for most server side operations. It is responsible
 * for registration and thread management for server protocols and channels,
 * dispatch and thread management for requests, object reference construction
 * and generates the adapter managers allowing flow control for adapters.
 *
 * @author <a href="mailto:wood@intalio.com">Chris Wood &lt;wood@intalio.com&gt;</a>
 * @version $Revision: 1.8 $ $Date: 2001/08/13 03:47:34 $ 
 */
public class ServerManagerImpl implements ServerManager {
  
  /**
   * Construct new server manager. There should be one of these objects per orb.
   * @param orb The controling orb.
   */
  public ServerManagerImpl(org.omg.CORBA.ORB orb) {
    _orb = orb;
    
    org.openorb.CORBA.kernel.ORBLoader loader = ((org.openorb.CORBA.ORB)orb).getLoader();

    // initialize policies.
    try {
      byte [] ssuid = loader.getStringProperty("openorb.server.alias", "").getBytes("UTF-8");
      _suid = new byte[ssuid.length+1];
      System.arraycopy(ssuid, 0, _suid, 0, ssuid.length);
      _suid[ssuid.length] = SEP_VAL;
    }
    catch(java.io.UnsupportedEncodingException ex) {
      throw new org.omg.CORBA.INITIALIZE(ex.toString());
    }
    
    _maxQueueSize = loader.getIntProperty("openorb.server.maxQueueSize", Integer.MAX_VALUE);
    _maxManagerHeldRequests = loader.getIntProperty("openorb.server.maxManagerHeldRequests", Integer.MAX_VALUE);
    _minThreadPoolSize = loader.getIntProperty("openorb.server.minThreadPoolSize", 1);
    _maxThreadPoolSize = loader.getIntProperty("openorb.server.maxThreadPoolSize", 200);
    
    _rootGroup = Thread.currentThread().getThreadGroup();

    long time = System.currentTimeMillis();
      
    _puid = new byte[8];
    _puid[7] = (byte)(time>>>56);
    _puid[6] = (byte)(time>>>48);
    _puid[5] = (byte)(time>>>40);
    _puid[4] = (byte)(time>>>32);
    _puid[3] = (byte)(time>>>24);
    _puid[2] = (byte)(time>>>16);
    _puid[1] = (byte)(time>>>8);
    _puid[0] = (byte)time;
    
    _reapCloseTime = loader.getIntProperty("openorb.server.reapCloseDelay", DEFAULT_CLOSE_TIME);
    _useStaticThreadGroup = loader.getBooleanProperty("openorb.useStaticThreadGroup", false);
  }
  
  // static state.
  
  /** 
   * Separates parts of object keys.
   */
  protected static final byte SEP_VAL = (byte)0xFE;
  
  /**
   * Escape character for allowing {@link SEP_VAL} as well as itself to be
   * inserted into object keys.
   */
  protected static final byte ESC_VAL = (byte)0xFD;
  
  /**
   * Procees unique identifier flag. This flag appears in the byte at index 3
   * to specify the id is process uneque rather than server uneque.
   */
  protected static final byte FLAG_PUID = 0x01;

  /**
   * This is a byte sequence which is uneque to this orb instance, the process
   * unique identifier. It is very unlikly that two orbs will share the same
   * puid, hostname, and listening port, thus nonpersistance of references
   * encapsulating this value is guaranteed. 
   */
  protected byte [] _puid;
  
  /**
   * This is a byte sequence which is uneque to this server, it can be
   * set with the -ORBServerAlias flag, or if unset defaults to an empty
   * sequence. Processes with the same suid can create persistent references
   * which are valid outside process boundaries.
   */
  private byte [] _suid;
  
  /**
   * Reference to the orb object.
   */
  private org.omg.CORBA.ORB _orb;
  
  /**
   * Thread group of thread which called the constructor.
   */
  private ThreadGroup _rootGroup;

  // exceptions for various request failure conditions.
  
  private static final org.omg.CORBA.SystemException QUEUE_FULL_EX 
    = new org.omg.CORBA.TRANSIENT(org.openorb.policy.OPENORB_VPVID.value | 1, CompletionStatus.COMPLETED_NO);
  private static final org.omg.CORBA.SystemException DISCARD_EX 
    = new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 1, CompletionStatus.COMPLETED_NO);
  private static final org.omg.CORBA.SystemException INACTIVE_EX 
    = new org.omg.CORBA.OBJ_ADAPTER(org.omg.CORBA.OMGVMCID.value | 1, CompletionStatus.COMPLETED_NO);
  private static final org.omg.CORBA.SystemException DESTROYED_EX 
    = new org.omg.CORBA.OBJECT_NOT_EXIST(0, CompletionStatus.COMPLETED_NO);
  private static final org.omg.CORBA.SystemException SHUTDOWN_EX 
    = new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 4, CompletionStatus.COMPLETED_NO);

  /**
   * @return orb associated with the server manager.
    */
  public org.omg.CORBA.ORB orb() {
    return _orb;
  }
  

  // channel and protocol registration.  
  
  /**
   * Synchronization for the protocol and channel sets
   */
  private final Object _syncIO = new Object();
  
  /**
   * Thread group containing IO threads.
   */
  private ThreadGroup _ioThreads;
  
  /**
   * Static thread group. Used in tests to avoid creating large numbers of 
   * empty thread groups.
   */
  private static ThreadGroup _staticIOThreads;

  /**
   * If true the static thread groups will be used in preference to the 
   * normal ones.
   */
  private boolean _useStaticThreadGroup;
  
  /**
   * Array of registered protocols and their profile IDs.
   */
  private Object [] _protProfIDs = new Object[0];
  
  /**
   * Active server protocols. This maps the protocol to the thread if
   * the protocol is listening, or to null if the protocol is not listening
   * but still requires a close on shutdown.
   */
  private Map _protToWorker = new HashMap();
  
  /**
   * Active server channels. This maps the channels to the worker threads
   * using the channel.
   */
  private Map _channels = new HashMap();
  
  /**
   * shuts down channels which have been quiet for a while
   */
  private Thread _channelReaper = null;

  /**
   * Minimum time for a channel to remain quiet before it's channel is
   * shut down. This is also used in thread pool size optimization.
   */
  private int _reapCloseTime;
  private static final int DEFAULT_CLOSE_TIME = 10*60*1000;

  // big scale lock
    
  /**
   * this syncronizes all the work queues, the states of the adapter
   * managers, and the adapter cache. A fairly course grained lock
   * although this avoids many deadlock and oversynchronization problems.
   */
  private Object _syncState = new Object();

  // state of the server manager synchronize both on _syncState and _syncIO
  // to modify.
  
  private volatile boolean _shutdown = false;
  private volatile boolean _running = false;  
  private volatile boolean _ioComplete = false;


  // adapter registration
  
  /**
   * This will be a special adapter which redirects all requests to a reference
   * resolved from resolve_initial_references.
   */
  private ObjectAdapter _defaultAdapter = null;
  
  /**
   * Map containing all adapters which have served a request.
   * In future this may be cleared out occasionaly of adapters which can 
   * be recreated easily.
   */
  private Map _adapterCache = new HashMap();
    
  // the request queues.
  
  /**
   * Total active requests, that is the number of requests which are
   * accessable via _requests, plus the number of requests which are
   * currently being processed.
   */
  private volatile int _activeRequests = 0;
  
  /**
   * Held requests. Requests waiting for some type of event. This counter
   * is kept more for checking than anything.
   */
  private volatile int _holdingRequests = 0;
  
  /**
   * Synchronization variable for thread pool and request queue. All 
   * variables below sync on this, up to misc.
   */
  private Object _syncQueue = new Object();
  
  /**
   * Requests queued ready for processing. All variables in this section
   * synchronize on this.
   */
  private MergeStack _requests = new MergeStack();  
  
  /**
   * Maximum value of _activeRequests for this time period. This value is reset
   * at the end of each time period to the average of the old _tpMaxActiveRequests
   * and the current value of _activeRequests.
   */
  private int _tpMaxActiveRequests = 0;
  
  /**
   * Value of _tpMaxActiveRequests for previous time period, used in calculating
   * _optThreadPoolSize.
   */
  private int _tpMaxActiveRequestsPrev = 0;  
  
  /**
   * Maximum queue length. This is the total number of requests which may be
   * queued before new incoming requests begin being rejected.
   */
  private int _maxQueueSize = 0;
  
  /**
   * Maximum number of manager held requests before new requests begin being
   * rejected.
   */
  private int _maxManagerHeldRequests = 0;
  
  /**
   * The current single threaded thread. There is only ever one.
   */
  private Thread _stThread = null;
  
  /**
   * Head of request queue for requests held waiting for single thread
   * availability.
   */
  private MergeStack _stRequests = new MergeStack();
  
  // thread pool for servicing work queue.

  /**
   * Current thread pool size. This can change dynamicaly.
   */
  private int _threadPoolSize = 0;
  
  /**
   * This set includes all server threads, both pool threads and threads
   * passed through serve_request.
   */
  private Set _serverThreads = new HashSet();
    
  /**
   * Allow use of thread pool.
   */
  private boolean _allowPool = false;
  
  /**
   * Thread group for thread pool threads
   */
  private ThreadGroup _poolThreads;

  /**
   * Static thread group. Used in tests to avoid creating large numbers of 
   * empty thread groups.
   */
  private static ThreadGroup _staticPoolThreads;
  
  /**
   * Array of pool threads.
   */
  private Thread [] _poolThreadArray = new Thread[10];
  
  /**
   * Maximum thread pool size. Thread pool will not grow beyond this size.
   */
  private int _maxThreadPoolSize = 5;
  
  /**
   * Minimum thread pool size. Thread pool will not shrink beyond this size.
   */
  private int _minThreadPoolSize = 1;
  
  /**
   * Optimum pool size. Based on the average of the maximum value of 
   * _activeRequests over the previous time period, and the maximum value
   * of _activeRequests over this time period.
   */
  private int _optThreadPoolSize = 0;
  
  // misc
  
  /**
   * Used in creating multicomponent profiles.
   */
  private org.omg.IOP.Codec _componentCodec;
  
  // policy settings
  
  /**
   * Set thread pool size ranges. The thread pool size will be somewhere 
   * within this range, depending on load.
   *
   * @param min minimum size for thread pool. Must be greater than 0 and less
   *         than max.
   * @param max maximum thread pool size. If this is 0 then requests will only
   *         be processed when the orb.perform_work function is called. Must be
   *         greater than 0.
   */
  public void setThreadPoolLimits(int min, int max) {
    if(min < 0 || max < 0 || min > max)
      throw new java.lang.IllegalArgumentException("Invalid thread pool size range");
    synchronized(_syncState) {
      _minThreadPoolSize = min;
      _maxThreadPoolSize = max;
    }
  }
  
  /**
   * Set maximum queue size. If the queue grows to this size then new incoming
   * requests will be rejected with a TRANSIENT exception. Changing this value
   * downwards will not result in queued requests being destroyed.
   *
   * @param maxQueueSize maximum queue size. Must be greater than 1. Values 
   *         should be several times the maximum thread pool size.
   */
  public void setMaxQueueSize(int maxQueueSize) {
    if(maxQueueSize < 1)
      throw new java.lang.IllegalArgumentException("Invalid max queue size");

    synchronized(_syncState) {
      _maxQueueSize = maxQueueSize;
    }
  }
  
  /**
   * Maximum number of requests which may be held by an adapter in the HOLDING
   * state. Changes to this value will only affect newly created object 
   * adapters.
   *
   * @param maxAdapterHeldRequests ceiling on adapter held requests. Individial
   *          adapter managers may modify this value downwards. Use 0 to 
   *          dissallow the adapter holding state, and Integer.MAX_VALUE for
   *          no limit.
   */
  public void setMaxManagerHeldRequests(int maxManagerHeldRequests) {
    if(maxManagerHeldRequests < 0)
      throw new java.lang.IllegalArgumentException("Invalid max held requests");
      
    synchronized(_syncState) {
      _maxManagerHeldRequests = maxManagerHeldRequests;
    }
  }
  
  // object id creation and manipulation.
  
  /**
   * This function is used by adapters to generate a key which will
   * participate in adapter lookup short-circuiting.
   */
  public byte [] create_cacheable_object_key(boolean use_suid, byte [][] parts)
  {
    
    // header + puid/suid + sep-chars
    int len = 4 + (use_suid?_suid.length:_puid.length) + parts.length-1;
    
    // count the number of bits needing escapes
    int escs = 0;
    for(int i = 0; i < parts.length; ++i) {
      len += parts[i].length;
      for(int j = 0; j < parts[i].length; ++j) {
        switch(parts[i][j]) {
          case SEP_VAL:
          case ESC_VAL:
            ++escs;
        }
      }
    }
    len += escs;
    byte [] object_key = new byte[len];
    
    // construct the header.
    object_key[0] = 0;
    object_key[1] = (byte)'O';
    object_key[2] = (byte)'O';
    int pos;
    if(use_suid) {
      object_key[3] = 0;
      System.arraycopy(_suid, 0, object_key, 4, _suid.length);
      pos = 4 + _suid.length;
    } else {
      object_key[3] = FLAG_PUID;
      System.arraycopy(_puid, 0, object_key, 4, _puid.length);
      pos = 4 + _puid.length;
    }
    
    int p = 0;
    while(p < parts.length && escs > 0) {
      for(int i = 0; i < parts[p].length; ++i, ++pos) {
        switch(parts[p][i]) {
          case SEP_VAL:
          case ESC_VAL:
            object_key[pos++] = ESC_VAL;
            escs--;
        }
        object_key[pos] = parts[p][i];
      }
      if(pos < object_key.length)
        object_key[pos++] = SEP_VAL;
      ++p;
    }
    for(;p < parts.length; ++p) {
      System.arraycopy(parts[p], 0, object_key, pos, parts[p].length);
      pos += parts[p].length;
      if(pos < object_key.length)
        object_key[pos++] = SEP_VAL;
    }
    return object_key;
  }
  
  /**
   * Extract the component parts of a cacheable object_key. Returns null if
   * is_cacheable_object_key would return false.
   */
  public byte [][] extract_cacheable_object_key(byte [] object_key) {
    int [] offsets = find_object_key_offsets(object_key);
    
    if(offsets == null)
      return null;
    
    byte [][] parts = new byte[offsets.length][];
    
    int end = object_key.length;
    for(int i = parts.length-1; i >= 0; --i) {
      int begin = offsets[i];
      
      // count the escapes
      int escs = 0;
      for(int j = begin; j < end; ++j)
        if(object_key[j] == ESC_VAL) {
          ++j;
          ++escs;
        }
      
      // extract the object_key part.
      parts[i] = new byte[end-begin-escs];
      if(escs == 0)
        System.arraycopy(object_key, begin, parts[i], 0, end-begin);
      else
        for(int j = begin, k = 0; j < end; ++j, ++k) {
          if(object_key[j] == ESC_VAL)
            ++j;
          parts[i][k] = object_key[j];
        }
      
      end = begin-1;
    }
    
    return parts;
  }
  
  private int [] find_object_key_offsets(byte [] object_key) {
    if(!is_local_cacheable_object_key(object_key))
      return null;
    int start = 4;
    
    if((object_key[3] & FLAG_PUID) == FLAG_PUID)
      start += _puid.length;
    else
      start += _suid.length;
    
    // count parts
    int parts = 1;
    for(int i = start; i < object_key.length; ++i)
      switch(object_key[i]) {
        case SEP_VAL:
          ++parts;
          break;
        case ESC_VAL:
          ++i;
          break;
      }
      
      // generate the offsets.
      int [] offsets = new int[parts];
      offsets[0] = start;
      int upto = 1;
      for(int i = start; i < object_key.length && upto < offsets.length; ++i)
        switch(object_key[i]) {
          case SEP_VAL:
            offsets[upto++] = i+1;
            break;
          case ESC_VAL:
            ++i;
            break;
        }
        
        return offsets;
  }
  
  /**
   * Returns true if the given object_key can be used in the cache.
   */
  public boolean is_cacheable_object_key(byte [] object_key) {
    return (object_key.length >= 4 && object_key[0] == 0 && 
            object_key[1] == (byte)'O' && object_key[2] == (byte)'O');
  }
  
  /**
   * Returns true if the given object key has a persistent target.
   */
  public boolean is_suid_object_key(byte [] object_key) {
    return (object_key.length >= 4 && (object_key[3] & FLAG_PUID) == 0);
  }
  
  /**
   * Returns true if the given object_key is cacheable and was
   * created by this server.
   */
  public boolean is_local_cacheable_object_key(byte [] object_key) {
    if(!is_cacheable_object_key(object_key))
      return false;
    
    if((object_key[3] & FLAG_PUID) == FLAG_PUID) {
      if(object_key.length - 4 < _puid.length)
        return false;
      for(int i = 0; i < _puid.length; ++i)
        if(object_key[i+4] != _puid[i])
          return false;
    } else {
      if(object_key.length - 4 < _suid.length)
        return false;
      for(int i = 0; i < _suid.length; ++i)
        if(object_key[i+4] != _suid[i])
          return false;
    }
    
    return true;
  }  
  
  // IOR management
  
  // TODO: all these functions should realy be located somewhere else.
  // This allows constructing IORs independently of the server side.
  
  /**
   * Register a server protocol. This registers the protocol for purposes of
   * constructing IORs. When the protocol begins listening it should call 
   * protocol_listening.
   *
   * If multiple protocols are registered with the same tag value, or a profile
   * is registered with multiple tag values then the constructed IOR will
   * contain multiple profiles, one from each distinct registered 
   * protocol/tag pair. Registering the same protocol with the same tag more
   * than once will not result in multiple profiles.
   *
   * @param profile_tag Profile tag of IOR consturcted with the protocol. 
   * @param protocol The server protocol.
   */
  public void register_protocol(int profile_tag, ServerProtocol protocol) {
    if(protocol == null)
      return;
    
    for(int i = 0; i < _protProfIDs.length; ++i)
      if(_protProfIDs[i] == protocol &&
         profile_tag == ((Integer)_protProfIDs[i+1]).intValue())
        return;
    
    Object [] tmp = new Object[_protProfIDs.length+2];
    System.arraycopy(_protProfIDs, 0, tmp, 0, _protProfIDs.length);
    tmp[_protProfIDs.length] = protocol;
    tmp[_protProfIDs.length+1] = new Integer(profile_tag);
    _protProfIDs = tmp;
  }
  
  /**
   * Construct an IOR.
   */
  public org.omg.IOP.IOR construct_ior(String type_id, byte[] object_key, org.openorb.PI.ComponentSet component_set) {
    org.omg.IOP.IOR ret = new org.omg.IOP.IOR();
    
    ret.type_id = type_id;
    
    int u = 0;
    
    org.omg.IOP.TaggedComponent [] multi = component_set.getComponents(org.omg.IOP.TAG_MULTIPLE_COMPONENTS.value);
    
    if(multi != null && multi.length > 0)
      ret.profiles = new org.omg.IOP.TaggedProfile[_protProfIDs.length/2+1];
    else {
      multi = null;
      ret.profiles = new org.omg.IOP.TaggedProfile[_protProfIDs.length/2];      
    }
    
    // add ordinary profiles.
    for(int i = 0; i < _protProfIDs.length; i+=2) {
      ServerProtocol protocol = (ServerProtocol)_protProfIDs[i];
      int profile_tag = ((Integer)_protProfIDs[i+1]).intValue();
      
      org.omg.IOP.TaggedProfile profile =
        protocol.create_profile(profile_tag, component_set, object_key);
      
      if(profile != null)
        ret.profiles[u++] = profile;
    }

    // if no active profiles we don't have a valid reference
    if(u == 0)
      return null;
    
    // add multicomponent profile.
    if(multi != null) {
      org.omg.CORBA.Any any = _orb.create_any();
      org.omg.IOP.TaggedComponentSeqHelper.insert(any, multi);
      
      org.omg.IOP.Codec cmptCodec = getComponentCodec();
      byte [] buf;
      try {
        buf = cmptCodec.encode_value(any);
      }
      catch(org.omg.IOP.CodecPackage.InvalidTypeForEncoding ex) {
        throw new org.omg.CORBA.INTERNAL(ex.toString());
      }
      
      ret.profiles[u++] = new org.omg.IOP.TaggedProfile(org.omg.IOP.TAG_MULTIPLE_COMPONENTS.value, buf);
    }
    
    if(u < ret.profiles.length) {
      org.omg.IOP.TaggedProfile [] tmp = new org.omg.IOP.TaggedProfile[u];
      System.arraycopy(ret.profiles, 0, tmp, 0, u);
      ret.profiles = tmp;
    }
    
    return ret;
  }
  
  private org.omg.IOP.Codec getComponentCodec() {
    if(_componentCodec == null) {
      try {
        org.omg.IOP.CodecFactory factory = (org.omg.IOP.CodecFactory)_orb.resolve_initial_references("CodecFactory");

        _componentCodec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)0));
      }
      catch(org.omg.CORBA.UserException ex) {
        throw new InternalError(ex.toString());
      }
    }
    return _componentCodec;
  }
      
  /**
   * Constructs a new IOR which is a merge of all the profiles in each of
   * the targets.
   */
  public org.omg.IOP.IOR merge_iors(String type_id, org.omg.IOP.IOR [] iors) {
    Set pms = new HashSet();
    for(int i = 0; i < iors.length; ++i)
      for(int j = 0; j < iors[i].profiles.length; ++j)
        pms.add(new ProfileMember(iors[i].profiles[j]));
    
    org.omg.IOP.IOR ret = new org.omg.IOP.IOR(type_id, new org.omg.IOP.TaggedProfile[pms.size()]);
    int i = 0;
    Iterator itt = pms.iterator();
    while(itt.hasNext())
      ret.profiles[i++] = ((ProfileMember)itt.next())._profile;
    return ret;
  }
  /**
   * Single use class, just hashes profiles.
   */
  private static class ProfileMember {
    ProfileMember(org.omg.IOP.TaggedProfile profile) {
      _profile = profile;
      _hash = _profile.tag;
      for (int i = 0; i < _profile.profile_data.length; ++i)
        _hash = 31*_hash + _profile.profile_data[i];
    }
    private int _hash;
    org.omg.IOP.TaggedProfile _profile;
    
    public int hashCode() {
      return _hash;
    }
    public boolean equals(Object obj) {
      if(!(obj instanceof ProfileMember))
        return false;
      ProfileMember pm2 = (ProfileMember)obj;
      if(_hash != pm2._hash || _profile.tag != pm2._profile.tag
      || _profile.profile_data.length != pm2._profile.profile_data.length)
        return false;
      for (int i = 0; i < _profile.profile_data.length; ++i)
        if(_profile.profile_data[i] != pm2._profile.profile_data[i])
          return false;
      return true;
    }
  }  
  
  // communication registration + thread management.
  
  /**
   * This is called when a server protocol begins listening. Thread resources
   * are allocated to call the listen function regularly.
   *
   * @param protocol The server protocol.
   */
  public void protocol_listening(ServerProtocol protocol) {    
    synchronized(_syncIO) {
      if(_shutdown)
        return;
      
      if(_ioThreads == null || _ioThreads.isDestroyed()) {
        if(_useStaticThreadGroup) { 
          if(_staticIOThreads == null || _staticIOThreads.isDestroyed())
            _staticIOThreads = new ThreadGroup(_rootGroup, "Server IO");
          _ioThreads = _staticIOThreads;
        } else
          _ioThreads = new ThreadGroup(_rootGroup, "Server IO");
      }
      
      // TODO: do somthing better than this. Have a thread pool for io.
      Thread p = new Thread(_ioThreads, new ProtocolRunner(protocol), "Worker for " + protocol.toString());
      p.setDaemon(true);
      p.start();
      
      _protToWorker.put(protocol, p);
    }    
  }
  private class ProtocolRunner implements Runnable {
    public ProtocolRunner(ServerProtocol svrproto) {
      _svrproto = svrproto;
    }
    private ServerProtocol _svrproto;
    public void run() {
      // this returns on interrupt or shutdown
      if(wait_for_startup())
        _svrproto.run_listen();
    }
  }
  
  /**
   * This is called when a server protocol is no longer listening. It stops
   * getting serviced by the pool thread.
   *
   * @param protocol The server protocol.
   * @param paused If true when the server side is shut down close will still
   *         be called on the protocol.
   */
  public void protocol_not_listening(ServerProtocol protocol, boolean paused) {
    Thread thread;
    synchronized(_syncIO) {
      if(paused) {
        thread = (Thread)_protToWorker.get(protocol);
        if(thread != null)
          _protToWorker.put(protocol, null);
      } else 
        thread = (Thread)_protToWorker.remove(protocol);
    }
    
    if(thread != null) {
      Thread curr = Thread.currentThread();
      thread.interrupt();
      while(thread != curr && thread.isAlive()) {
        try {
          thread.join();
        }
        catch(InterruptedException ex) {}
      }
    }
  }

  /**
   * This operation occours when a protocol accepts a new incoming channel.
   */
  public void register_channel(ServerChannel channel) {
    synchronized(_syncIO) {
      if(_shutdown)
        return;
      
      if(_channels.containsKey(channel))
        return;
      
      if(_ioThreads == null || _ioThreads.isDestroyed()) {
        if(_useStaticThreadGroup) { 
          if(_staticIOThreads == null || _staticIOThreads.isDestroyed())
            _staticIOThreads = new ThreadGroup(_rootGroup, "Server IO");
          _ioThreads = _staticIOThreads;
        } else
          _ioThreads = new ThreadGroup(_rootGroup, "Server IO");
      }
      
      Thread thread = new Thread(_ioThreads, new ChannelRecvRunner(channel), "Recieve Worker for " + channel.toString());
      thread.setDaemon(true);
      thread.start();

      _channels.put(channel, thread);
    }    
  }
  private class ChannelRecvRunner implements Runnable {
    public ChannelRecvRunner(ServerChannel chan) {
      _chan = chan;
    }
    private ServerChannel _chan;
    public void run() {
      // this returns on interrupt or shutdown
      if(wait_for_startup())
        _chan.run_recv();
    }
  }
  
  /**
   * Called when a channel enters the closed state. Returns once
   * all channel threads have completed their work cycles.
   */
  public void unregister_channel(ServerChannel channel) {
    Thread thread;
    synchronized(_syncIO) {
      thread = (Thread) _channels.remove(channel);
    }
    
    if(thread != null) {
      Thread curr = Thread.currentThread();
      thread.interrupt();
      while(thread != curr && thread.isAlive()) {
        try {
          thread.join();
        }
        catch(InterruptedException ex) {}
      }
    }
  }

  /**
   * This fuction runs in a separate thread and cleans up 
   * any channels which are not used for a long time.
   */
  public void channel_reaper() {
    int lasttime = RequestIDAllocator.get_request_id();

    ServerChannel channel;
    Iterator itt;
    ServerChannel [] chans = new ServerChannel[0];
    while(!_shutdown) {
      try {
        Thread.sleep(_reapCloseTime);
      }
      catch(InterruptedException ex) {
        return;
      }
      
      if(_shutdown)
        return;
      
      synchronized(_syncQueue) {
        _tpMaxActiveRequestsPrev = _tpMaxActiveRequests;
        _tpMaxActiveRequests = _activeRequests;
        recalcOptThreadPoolSize();
        if(_optThreadPoolSize < _threadPoolSize)
          _syncQueue.notifyAll();
      }
      
      synchronized(_syncIO) {
        if(_channels.isEmpty())
          continue;
        chans = (ServerChannel[])_channels.keySet().toArray(chans);
      }
      
      for(int i = 0; i < chans.length && chans[i] != null; ++i)
        if(chans[i].channel_age() < lasttime)
          chans[i].soft_close(false);

      lasttime = RequestIDAllocator.get_request_id();
    }
  }

  // adapter location, registration and unregistration.
  
  /**
   * This function finds an adapter for the specified object id. If an adapter
   * for an object with the given object id cannot be found this returns null.
   * This varient of the find_adapter function is used by local bindings. It 
   * ignores the state of the adapter manager and any single threading policies,
   * If the target adapter or one of it's ancestors is in the process of being
   * destroyed this will wait until the destruction is complete.
   *
   * @throws AdapterDestroyedException The target adapter is in the process
   *          of being destroyed. Requests for the adapter may be retried once
   *          the destruction of the adapter is complete. 
   * @return The adapter specified in the object key.
   */
  public ObjectAdapter find_adapter(byte [] object_key) {
    return find_adapter(object_key, null);
  }
  
  /**
   * This function finds an adapter for the specified object id. If an adapter
   * for an object with the given object id cannot be found this returns null.
   * This varient of the find_adapter function is used by remote bindings. It 
   * takes into account the state of any adapter managers when accessing user
   * code, throwing an AdapterHoldingException if they are in the holding state.
   *
   * @param object_key the object key from the target.
   * @param The server request. This will be null for local invocations.
   * @throws AdapterDestroyedException The target adapter is in the process
   *          of being destroyed. Requests for the adapter may be retried once
   *          the destruction of the adapter is complete. 
   * @throws AdapterHoldingException User code must be entered to create the
   *          target adapter and the adapter manager is in the holding state.
   *          The request is must be until the holding state is exited.
   */
  private ObjectAdapter find_adapter(byte [] object_key, ServerRequest request) {
    // see if the object_key refers to a cacheable object.
    if(!is_cacheable_object_key(object_key))
      return _defaultAdapter;
    if(!is_local_cacheable_object_key(object_key)) {
      // cancel the request.
      if(request != null)
        request.server_cancel(new org.omg.CORBA.OBJECT_NOT_EXIST(org.omg.CORBA.OMGVMCID.value | 2, org.omg.CORBA.CompletionStatus.COMPLETED_NO));
      return null;
    }
    
    int [] offsets = find_object_key_offsets(object_key);
    ProgressiveAIDKey aid_key = new ProgressiveAIDKey(object_key, offsets);

    while(true) {
      // see if the adapter, or one of it's parents is in the cache.
      // since cache misses are relitivly fast to calculate, we try from most
      // specific to most general.
      AdapterValue ada_val = null;
      synchronized(_syncState) {
        aid_key.useHash(offsets.length-1);
        ada_val = (AdapterValue)_adapterCache.get(aid_key);
        if(ada_val != null) {
          // direct target, return it without messing around
          ++(ada_val._hits);
          return ada_val._adapter;
        }
        for(int i = offsets.length-2; i >= 0 && ada_val == null; --i) {
          aid_key.useHash(i);
          ada_val = (AdapterValue)_adapterCache.get(aid_key);
        }
      }
      if(ada_val == null) {
        // cancel the request.
        if(request != null)
          request.server_cancel(new org.omg.CORBA.OBJECT_NOT_EXIST(org.omg.CORBA.OMGVMCID.value | 2, org.omg.CORBA.CompletionStatus.COMPLETED_NO));
        return null;
      }

      ObjectAdapter adapter = ada_val._adapter;

      // check for direct child
      if(adapter.object_id(object_key) != null) {
        ++(ada_val._hits);
        return adapter;
      }

      boolean is_st_thread = false;
      boolean made_st_thread = false;

      // user code must be entered to create an adapter.
      try {
        while(true) {
          // check the adapter's state.
          AdapterManagerImpl manager = null;
          if(request != null) {
            if(!is_st_thread && adapter.single_threaded())
              synchronized(_syncQueue) {
                if(_stThread == Thread.currentThread())
                  is_st_thread = true;
                else if(_stThread == null) {
                  _stThread = Thread.currentThread();
                  made_st_thread = true;
                  is_st_thread = true;
                } else {
                  // queue request in st queue
                  _stRequests.addLast(request);
                  ++_holdingRequests;
                  return null;
                }
              }

            manager = (AdapterManagerImpl)adapter.getAdapterManager();

            if(manager != null && !manager.begin_request(request))
              return null;
          }

          // find a child adapter.
          try {
            adapter = adapter.find_adapter(object_key);
          }
          catch(AdapterDestroyedException ex) {
            // one of the adapters in the chain has been destroyed.
            byte [] aid = ex.getAdapterID();

            int [] aoff = find_object_key_offsets(aid);
            int offset = aoff[aoff.length-1];

            AdapterValue child_ada_val 
              = new AdapterValue(aid, offset, ex.getObjectAdapter());
            synchronized(_syncState) {
              child_ada_val = (AdapterValue)_adapterCache.get(child_ada_val);
              // child adapter already destroyed.
              // run the whole method again from the start
              if(child_ada_val == null)
                break;

              if(request != null) {
                // hold request and wait for adapter to be detroyed.
                if(child_ada_val._destrRequests == null)
                  child_ada_val._destrRequests = new MergeStack();
                child_ada_val._destrRequests.addLast(request);
                ++_holdingRequests;
                return null;
              } else {
                // local request, just wait, don't hold.
                child_ada_val.waiting_for_destroy = true;
                boolean interrupted = false;
                while(child_ada_val.waiting_for_destroy) {
                  try {
                    _syncState.wait();
                  }
                  catch(InterruptedException iex) {
                    interrupted = true;
                  }
                }
                if(interrupted)
                  Thread.currentThread().interrupt();
                // we need to start the entire process from the beginning
                // the current adapter may be the one which has been destroyed.
                break;
              }
            }
          }
          finally {
            if(manager != null)
              manager.complete_request(request);
          }

          if(adapter == null) {
            // cancel the request.
            if(request != null)
              request.server_cancel(new org.omg.CORBA.OBJECT_NOT_EXIST(org.omg.CORBA.OMGVMCID.value | 2, org.omg.CORBA.CompletionStatus.COMPLETED_NO));
            return null;
          }

          if(adapter.object_id(object_key) != null)
            return adapter;
        }
      }
      finally {
        if(made_st_thread)
          synchronized(_syncQueue) {
            _stThread = null;
          }
      }
    }
  }  
  
  /**
   * Register an adapter. If the specified adapter is already
   * registered no change occours. This will occour automaticaly for
   * child adapters of some root adapter.
   */
  public void register_adapter(byte [] aid, ObjectAdapter adapter) {
    if(aid.length == 0) {
      // default adapter.
      if(_defaultAdapter == null)
        _defaultAdapter = adapter;
      return;
    }
    
    if(!is_local_cacheable_object_key(aid))
      throw new org.omg.CORBA.BAD_PARAM(0, CompletionStatus.COMPLETED_NO);
    
    int offset = (find_object_key_offsets(aid))[0];
    
    AdapterValue ada_val = new AdapterValue(aid, offset, adapter);
    AdapterManagerImpl manager = (AdapterManagerImpl)adapter.getAdapterManager();
    
    synchronized(_syncState) {
      if(_adapterCache.containsKey(ada_val))
        return;
      
      if(manager != null)
        manager.addManagedAdapter(adapter);
      
      _adapterCache.put(ada_val, ada_val);
    }
  }
  
  /**
   * Unregister an adapter. If the specified adapter is not registered
   * no change occours. If an adapter throws an AdapterDestroyed exception
   * it must eventualy call this function to continue processing the held 
   * requests.
   */
  public void unregister_adapter(byte [] aid) {
    if(aid.length == 0) {
      _defaultAdapter = null;
      return;
    }
    
    if(!is_local_cacheable_object_key(aid))
      throw new org.omg.CORBA.BAD_PARAM(0, CompletionStatus.COMPLETED_NO);
    
    int offset = (find_object_key_offsets(aid))[0];
    AdapterValue ada_val = new AdapterValue(aid, offset, null);
    
    synchronized(_syncState) {
      ada_val = (AdapterValue)_adapterCache.remove(ada_val);
      if(ada_val == null)
        return;
      if(ada_val.waiting_for_destroy) {
        ada_val.waiting_for_destroy = false;
        _syncState.notifyAll();
      }
    }

    // unregister the adapter from it's manager
    AdapterManagerImpl manager = (AdapterManagerImpl)ada_val._adapter.getAdapterManager();

    if(manager != null)
      manager.removeManagedAdapter(ada_val._adapter);
    
    // return any requests blocked waiting for destruction to the queue.
    if(ada_val._destrRequests != null && !ada_val._destrRequests.isEmpty()) {
      synchronized(_syncQueue) {
        _activeRequests += ada_val._destrRequests.size();
        _holdingRequests -= ada_val._destrRequests.size();
        _requests.append(ada_val._destrRequests);
        _syncQueue.notifyAll();
      }
    }
  }
  
  /**
   * Create an adapter manager.
   */
  public ServerManager.AdapterManager create_adapter_manager() {
    return new AdapterManagerImpl();
  }  
  
  // queue management
  
  /**
   * Begin dispatch sequence for incoming request. This will enqueue the
   * request and return immediatly.
   */
  public void enqueue_request(ServerRequest request) {
    // TODO: add per-client queue size checks.        
    
    if(_shutdown)
      request.server_cancel(SHUTDOWN_EX);
    
    synchronized(_syncQueue) {
      if(_maxQueueSize == 0 || _activeRequests < _maxQueueSize) {
        ++_activeRequests;
        _requests.addLast(request);
        _syncQueue.notify();        
        
        // update the optimum thread pool size        
        if(_activeRequests > _tpMaxActiveRequests) {
          _tpMaxActiveRequests = _activeRequests;
          recalcOptThreadPoolSize();
        }
        
        // increace the size of the thread pool.
        if(_threadPoolSize < _minThreadPoolSize) 
          extend_thread_pool(_minThreadPoolSize-_threadPoolSize);
        else if(_threadPoolSize < _maxThreadPoolSize && _activeRequests > _threadPoolSize)
          extend_thread_pool(1);
        
        return;
      }
    }
    
    request.server_cancel(QUEUE_FULL_EX);
  }

  /**
   * Calculate optimum thread pool size.
   */
  private void recalcOptThreadPoolSize() {
    if(_tpMaxActiveRequestsPrev > _tpMaxActiveRequests)
      _optThreadPoolSize = (_tpMaxActiveRequests + _tpMaxActiveRequestsPrev) / 2;
    else
      _optThreadPoolSize = _tpMaxActiveRequests * 11 / 10;
  }
  
  public boolean work_pending() {
    if(_shutdown)
      return false;
    if(!_running) {        
      synchronized(_syncIO) {
        return !_channels.isEmpty();
      }
    }
    synchronized(_syncQueue) {
      return !_requests.isEmpty();
    }
  }

  /**
   * If this value is passed to the next_request function 
   * the function will serve a single request, waiting if one is
   * not available on the queue, and return.
   */
  private static final int POOL_ID_WAIT = -1;
  
  /**
   * If this value is passed to the next_request function 
   * the function will serve a single request, not waiting if one is
   * not available on the queue, and return.
   */  
  private static final int POOL_ID_NO_WAIT = -2;
  
  /**
   * This is called from orb.perform_work, it serves a single request.
   *
   * @returns true if a request was served, false if the queue is empty
   */
  public boolean serve_request(boolean wait) {
    if(!_running)
      startup(false, false);
    
    // main function will deal with cleaning up afterwards.
    return thread_pool_main(wait?POOL_ID_WAIT:POOL_ID_NO_WAIT);
  }

  /**
   * Main function used by thread pool threads and also by the 
   * serve_request function.
   */
  private boolean thread_pool_main(int pool_id) {
    // may retain st flag across loops.
    Thread currThread = Thread.currentThread();
    if(pool_id < 0)
      synchronized(_syncQueue) {
        _serverThreads.add(currThread);
      }
  
    boolean completedRequest = false;
    while(true) {
      ServerRequest nextRequest = null;      
      synchronized(_syncQueue) {
        while(true) {
          // decrement active requests. Whatever happened, the last loop
          // successfully completed a request.
          if(completedRequest)
            --_activeRequests;

          // thread pool shutdown 'function'
          if(_shutdown || (completedRequest && pool_id < 0) 
            || (pool_id >= _optThreadPoolSize && pool_id >= _minThreadPoolSize)) {          
            // remove thread from pool
            if(pool_id >= 0) {
              _poolThreadArray[pool_id] = null;
              _threadPoolSize--;
              org.openorb.util.Trace.high("ServerManager", "Thread pool shrunk to " + _threadPoolSize);
            }

            // unmark this as the single thread thread.
            if(_stThread == currThread)
              _stThread = null;

            // notify shutdown that no more pool threads are active
            _serverThreads.remove(currThread);  
            if(_shutdown && _serverThreads.isEmpty())
              _syncQueue.notifyAll();

            return completedRequest;
          }

          completedRequest = false;  

          // get the first available request.
          if(!_stRequests.isEmpty() && (_stThread == null || _stThread == currThread)) {
            if(_stThread != currThread)
              _stThread = currThread;
            nextRequest = (ServerRequest)_stRequests.removeFirst();
            ++_activeRequests;
            --_holdingRequests;
          } else {
            // lose the st flag. We're getting req from the incomming
            if(_stThread == currThread)
              _stThread = null;

            // wait for a request to arrive
            if(_requests.isEmpty()) {            
              if(pool_id != POOL_ID_NO_WAIT) {
                try {
                  _syncQueue.wait();
                  continue;
                }
                catch(InterruptedException ex) {
                  Thread.currentThread().interrupt();

                  _serverThreads.remove(currThread);
                  if(_shutdown && _serverThreads.isEmpty())
                    _syncQueue.notifyAll();
                  return false;
                }
              }            
            }

            nextRequest = (ServerRequest)_requests.removeFirst();    
          }
          break;
        }
      }
      
      serve_request(nextRequest);
      
      completedRequest = true;
    }
  }
    
  /**
   * This function is responsible for all the queue management, it takes
   * an incoming request from the head of the queue and either dispatches
   * the request or puts it in the appropriate holding location.
   */
  private void serve_request(ServerRequest request) {
    int state = request.begin_request();
    
    while(true) {
      // find the adapter for the request.
      ObjectAdapter adapter;
      switch(state) {
      case ServerRequest.STATE_FIND_ADAPTER:
        adapter = find_adapter(request.object_key(), request);

        if(adapter == null ||
          (state = request.adapter(adapter)) == ServerRequest.STATE_COMPLETE) {
          return;
        }
        break;
      case ServerRequest.STATE_COMPLETE:
        return;
      default:
        adapter = request.adapter();
      }

      if(state != ServerRequest.STATE_QUEUED)
        throw new InternalError();

      // deal with single threaded adapters.
      if(adapter.single_threaded()) {
        synchronized(_syncQueue) {
          if(_stThread == null)
            _stThread = Thread.currentThread();
          else if(_stThread != Thread.currentThread()) {
            // queue request in st queue
            _stRequests.addLast(request);
            ++_holdingRequests;
            return;
          }
        }
      }

      // see if the manager rejects or queues the request.
      AdapterManagerImpl manager = (AdapterManagerImpl)adapter.getAdapterManager();

      if(manager != null && !manager.begin_request(request))
        return;

      try {
        // this may throw an AdapterDestroyed exception
        request.dispatch();
      }
      catch(AdapterDestroyedException ex) {
        // the adapter has been destroyed. If there is still an entry in
        // the adapter lookup table store the request with it for re-finding
        // when the target is eventualy destroyed,
        byte [] aid = ex.getAdapterID();
        org.openorb.util.Trace.assertTrue(adapter == ex.getObjectAdapter());

        int [] aid_offsets = find_object_key_offsets(aid);

        AdapterValue ada_val = new AdapterValue(aid, aid_offsets[0], adapter);
        synchronized(_syncState) {
          ada_val = (AdapterValue)_adapterCache.get(ada_val);

          if(ada_val != null && ada_val._adapter == adapter) {
            // hold request and wait for adapter to be detroyed.
            if(ada_val._destrRequests == null)
              ada_val._destrRequests = new MergeStack();
            ada_val._destrRequests.addLast(request);
            ++_holdingRequests;
            return;
          }
        }

        // the adapter has been destroyed try again. (the request may
        // result in a new adapter being created).
        continue;
      }
      finally {
        if(manager != null)
          manager.complete_request(request);
      }

      return;
    }
  }
  
  /**
   * Determine if this thread is a server thread.
   */
  private boolean is_server_thread() {
    synchronized(_syncState) {
      return _serverThreads.contains(Thread.currentThread());
    }
  }

  /**
   * Extend the thread pool with one more thread. This will be called
   * from the enqueue request function when the thread pool must grow.
   * Lock on sync_state must be held.
   */
  private void extend_thread_pool(int count) {
    if(_poolThreads == null || _poolThreads.isDestroyed()) {
      if(_useStaticThreadGroup) { 
        if(_staticPoolThreads == null || _staticPoolThreads.isDestroyed())
          _staticPoolThreads = new ThreadGroup(_rootGroup, "Server Threads");
        _poolThreads = _staticPoolThreads;
      } else
        _poolThreads = new ThreadGroup(_rootGroup, "Server Threads");
    }
        
    if(_poolThreadArray.length < _threadPoolSize+count) {
      Thread [] tmp = new Thread[_threadPoolSize+count];
      System.arraycopy(_poolThreadArray, 0, tmp, 0, _poolThreadArray.length);
      _poolThreadArray = tmp;
    }
    _threadPoolSize += count;
    
    int id = 0;
    for(int i = 0; i < count; ++i) {
      while(_poolThreadArray[id] != null)
        ++id;
       
      _poolThreadArray[id] = new PoolThread(_poolThreads, id);
      _serverThreads.add(_poolThreadArray[id]);
      _poolThreadArray[id].setDaemon(true);
      _poolThreadArray[id].start();
      ++id;
      
      org.openorb.util.Trace.high("ServerManager", "Thread pool grew to " + _threadPoolSize);
    }
  }
  private class PoolThread extends Thread {
    private PoolThread(ThreadGroup group, int id) {
      super(group, "Thread pool thread #" + id);
      _id = id;
    }
    private int _id;
    public void run() {
      thread_pool_main(_id);
    }
  }
  
  private boolean wait_for_startup() {
    if(!_running) {
      synchronized(_syncState) {
        while(!_running) {
          if(_shutdown)
            return false;
          try {
            _syncState.wait();
          }
          catch(InterruptedException ex) {
            return false;
          }
        }
      }
    }
    return true;
  }
  
  public void startup(boolean block, boolean allowPool) {
    synchronized(_syncState) {
      if(_shutdown)
        return;
      
      
      if(!_running) {
        _running = true;
        
        _allowPool = allowPool;
        
        // start the channel reaper.
        if(_ioThreads == null || _ioThreads.isDestroyed()) {
          if(_useStaticThreadGroup) { 
            if(_staticIOThreads == null || _staticIOThreads.isDestroyed())
              _staticIOThreads = new ThreadGroup(_rootGroup, "Server IO");
            _ioThreads = _staticIOThreads;
          } else
            _ioThreads = new ThreadGroup(_rootGroup, "Server IO");
        }
        
        if(_reapCloseTime > 0) {
          _channelReaper = new Thread(_ioThreads, new Runnable() {
              public void run() { channel_reaper(); }
            }, "Channel Reaper");
          _channelReaper.setDaemon(true);
          _channelReaper.start();
        }
        
        // start all the waiting threads.
        _syncState.notifyAll();        
      }  else if(allowPool && !_allowPool) {
        // thread pool now allowed.
        _allowPool = true;
        if(_optThreadPoolSize<_minThreadPoolSize)
          extend_thread_pool(_optThreadPoolSize);
        else if(_optThreadPoolSize<_maxThreadPoolSize)
          extend_thread_pool(_optThreadPoolSize);
        else
          extend_thread_pool(_maxThreadPoolSize);
      }
          
        
      if(block) {
        try {
          while(_running)
            _syncState.wait();
        }
        catch(InterruptedException ex) {
          // allow the server thread to be interrupted.
        }
      }
    }
  }
  
  public void shutdown(boolean wait_for_complete) {
    // spawn a thread to do the shutdown if wait for complete is false
    if(!wait_for_complete) {
      synchronized(_syncState) {
        if(_shutdown)
          return;
        Thread t = new Thread(_rootGroup, new Runnable() {
          public void run() {
            shutdown(true);
          }
        }, "ORB shutdown thread");
        t.setDaemon(true);
        t.start();
        while(!_ioComplete) {
          try {
            _syncState.wait();
          }
          catch(InterruptedException ex) {
            Thread.currentThread().interrupt();
          }
        }
      }
      return;
    }
    
    // make sure we aren't calling this with a server thread.
    if(is_server_thread())
      throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE);
    
    // stop listening for incoming connection requests + change to shutdown
    // state
    boolean interrupt = false;
    ServerProtocol [] protocols;
    synchronized(_syncState) {
      synchronized(_syncIO) {
        if(_shutdown)
          return;
        _shutdown = true;
        
        protocols = new ServerProtocol[_protToWorker.size()];
        protocols = (ServerProtocol [])_protToWorker.keySet().toArray(protocols);
      }

      // stop the channel reaper
      if(_channelReaper != null)
        _channelReaper.interrupt();
    }
        
    // close all the protocols.
    for(int i = 0; i < protocols.length; ++i)
      if(protocols[i] != null)
        protocols[i].close();
    
    synchronized(_syncState) {
      // allow any threads blocked in shutdown(false) to return.
      _ioComplete = true;
      _syncState.notifyAll();
    }
      
    synchronized(_syncQueue) {
      // notify server threads to wake up and notice death.
      _syncQueue.notifyAll();
      
      // wait for all server threads to complete.
      while(!_serverThreads.isEmpty()) {
        try {
          _syncQueue.wait();
        }
        catch(InterruptedException ex) {
          interrupt = true;
        }
      }
    }
      
    synchronized(_syncState) {
      // cancel any queued or held requests.
      while(!_stRequests.isEmpty()) {
        ServerRequest request = (ServerRequest)_stRequests.removeFirst();
        request.server_cancel(SHUTDOWN_EX);
        --_holdingRequests;
      }
            
      while(!_requests.isEmpty()) {
        ServerRequest request = (ServerRequest)_requests.removeFirst();
        request.server_cancel(SHUTDOWN_EX);
        --_activeRequests;
      }

      // etherealize adapters and cancel adapter held requests
      Iterator itt = _adapterCache.values().iterator();
      while(itt.hasNext()) {
        AdapterValue ada_val = (AdapterValue)itt.next();

        // discard any the held requests
        AdapterManagerImpl manager = (AdapterManagerImpl)ada_val._adapter.getAdapterManager();
        if(manager != null && manager._state == State.HOLDING) {
          if(manager._holdRequests != null)
            while(!manager._holdRequests.isEmpty()) {
              ServerRequest request = (ServerRequest)manager._holdRequests.removeFirst();
              request.server_cancel(SHUTDOWN_EX);
              --_holdingRequests;
            }

          manager._state = State.INACTIVE;
        }

        // etherealize the adapter.
        ada_val._adapter.etherealize(true);
      }

      // do a few tests...
      org.openorb.util.Trace.assertTrue(_activeRequests == 0, "Active requests not empty");
      org.openorb.util.Trace.assertTrue(_holdingRequests == 0, "Holding requests not empty");
    }
    
    // wait for channel reaper to complete.
    if(_channelReaper != null) {
      try {
        _channelReaper.join();
        _channelReaper = null;
      }
      catch(InterruptedException ex) {
        interrupt = true;
      }
    }
    
    // close the server channels.
    ServerChannel [] channels;
    synchronized(_syncIO) {
      channels = new ServerChannel[_channels.size()];
      channels = (ServerChannel[])_channels.keySet().toArray(channels);
    }
    for(int i = 0; i < channels.length; ++i)
      channels[i].soft_close(true);

    // allow anything blocked in startup to complete
    synchronized(_syncState) {
      _running = false;
      _syncState.notifyAll();
    }

    // destroy thread groups.
    if(!_useStaticThreadGroup) {
      if(_ioThreads != null) {
        _ioThreads.setDaemon(true);
        try {
          _ioThreads.destroy();
        }
        catch(IllegalThreadStateException ex) {}
      }
      if(_poolThreads != null) {
        _poolThreads.setDaemon(true);
        try {
          _poolThreads.destroy();
        }
        catch(IllegalThreadStateException ex) {}
      }
    }
    
    if(interrupt)
      Thread.currentThread().interrupt();
  }
  
  // adapter manager class
  
  public class AdapterManagerImpl
    extends org.omg.CORBA.LocalObject
    implements org.omg.PortableServer.POAManager, ServerManager.AdapterManager
  {
    /**
     * This class can only be instantiated by a server manager.
     */
    AdapterManagerImpl() {
    }
    
    private Set _managedAdapters = new HashSet();
    
    private State _state = State.HOLDING;
    
    private int _managedActiveRequests = 0;
    
    private MergeStack _holdRequests;

    private boolean _etherealize;

    private int _maxManagerHeldRequestsOverride = _maxManagerHeldRequests;
    
    public void setMaxManagerHeldRequests(int max) {
      if(max < 0)
        throw new IllegalArgumentException();
      
      _maxManagerHeldRequestsOverride = max;
    }

    synchronized public org.omg.PortableServer.POAManagerPackage.State get_state()
    {
      return _state;
    }
  
    public void deactivate(boolean etherealize, boolean wait_for_completion)
      throws org.omg.PortableServer.POAManagerPackage.AdapterInactive
    {
      // test to ensure this thread is not a server thread.
      if(wait_for_completion && is_server_thread())
        throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 6, CompletionStatus.COMPLETED_NO);

      synchronized(this) {
        if(_state == State.INACTIVE) {
          if(_managedActiveRequests == 0)
            throw new org.omg.PortableServer.POAManagerPackage.AdapterInactive();
          if(wait_for_completion)
            wait_for_complete();
          return;
        }
        _state = State.INACTIVE;
        if(!wait_for_completion && etherealize && _managedActiveRequests > 0) {
          _etherealize = true;
          etherealize = false;
        }
        if(wait_for_completion && _managedActiveRequests == 0)
          wait_for_completion = false;
      }
      
      if(_holdRequests != null) {
        while(!_holdRequests.isEmpty()) {
          ServerRequest request = (ServerRequest)_holdRequests.removeFirst();
          request.server_cancel(INACTIVE_EX);
        }
        _holdingRequests -= _holdRequests.size();
      }

      if(wait_for_completion || etherealize)
        synchronized(this) {
          if(wait_for_completion)
            wait_for_complete();
          if(etherealize)
            complete_deactivate();
        }
    }
    
    private void complete_deactivate() {
      ObjectAdapter [] adapters = new ObjectAdapter[_managedAdapters.size()];
      _managedAdapters.toArray(adapters);
      for(int i = 0; i < adapters.length; ++i)
        adapters[i].etherealize(true);
    }
    
    private void wait_for_complete() {
      try {
        while(_managedActiveRequests > 0 && _state != State.ACTIVE)
          wait();
      }
      catch(InterruptedException ex) {
        // we aren't in a server thread, just return.
        return;
      }
    }

    public void discard_requests(boolean wait_for_completion)
      throws org.omg.PortableServer.POAManagerPackage.AdapterInactive
    {
      // test to ensure this thread is not a server thread.
      if(wait_for_completion && is_server_thread())
        throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 6, CompletionStatus.COMPLETED_NO);
      
      MergeStack hold = null;
      synchronized(this) {
        switch(_state.value()) {
        case State._INACTIVE:
          throw new org.omg.PortableServer.POAManagerPackage.AdapterInactive();
        case State._HOLDING:
          if(_holdRequests != null && !_holdRequests.isEmpty()) {
            hold = _holdRequests;
            _holdRequests = null;
          }
        }
        _state = State.DISCARDING;
        if(hold == null || hold.isEmpty()) {
          if(wait_for_completion)
            wait_for_complete();
          return;
        }
      }
      
      // discard all the previously held requests
      while(!hold.isEmpty()) {
        ServerRequest request = (ServerRequest)hold.removeFirst();
        request.server_cancel(DISCARD_EX);
        --_holdingRequests;
      }
      
      if(wait_for_completion) {
        synchronized(this) {
          wait_for_complete();
        }
      }
    }
    
    public void hold_requests(boolean wait_for_completion)
      throws org.omg.PortableServer.POAManagerPackage.AdapterInactive
    {
      // test to ensure this thread is not a server thread.
      if(wait_for_completion && is_server_thread())
        throw new org.omg.CORBA.BAD_INV_ORDER(org.omg.CORBA.OMGVMCID.value | 6, CompletionStatus.COMPLETED_NO);
  
      synchronized(this) {
        if(_state == State.INACTIVE)
          throw new org.omg.PortableServer.POAManagerPackage.AdapterInactive();
        _state = State.HOLDING;
        
        if(wait_for_completion)
          wait_for_complete();
      }
    }
  
    public void activate()
      throws org.omg.PortableServer.POAManagerPackage.AdapterInactive
    {
      MergeStack hold = null;
      synchronized(this) {
        switch(_state.value()) {
        case State._INACTIVE:
          throw new org.omg.PortableServer.POAManagerPackage.AdapterInactive();
        case State._ACTIVE:
          return;
        case State._HOLDING:
          hold = _holdRequests;
          _holdRequests = null;
        }
        _state = State.ACTIVE;
        notifyAll();
      }

      if(hold != null && !hold.isEmpty()) {
        synchronized(_syncQueue) {      
          _holdingRequests -= hold.size();
          _activeRequests += hold.size();
          _requests.append(hold);
          _syncQueue.notifyAll();
        }
      }
    }
    
    boolean begin_request(ServerRequest request) {
      org.omg.CORBA.SystemException respond_ex = null;
      synchronized(this) {
        switch(_state.value()) {
        case State._ACTIVE:
          ++(_managedActiveRequests);
          return true;
        case State._HOLDING:
          if(_holdRequests == null || _holdRequests.size() < 
             ((_maxManagerHeldRequestsOverride<_maxManagerHeldRequests)
               ?_maxManagerHeldRequestsOverride:_maxManagerHeldRequests)) {

            if(_holdRequests == null)
              _holdRequests = new MergeStack();
            _holdRequests.addLast(request);

            ++_holdingRequests;
            return false;
          }
          // fallthrough. If max holding limit is reached treat like
          // the discarding state.
        case State._DISCARDING:
          respond_ex = DISCARD_EX;
          break;
        case State._INACTIVE:
          respond_ex = INACTIVE_EX;
          break;
        }
      }
      request.server_cancel(respond_ex);
      return false;
    }
      
    void complete_request(ServerRequest request) {
      synchronized(this) {
        if(--_managedActiveRequests == 0) {
          if(_etherealize)
            complete_deactivate();
          notifyAll();
        }
      }
    }
    
    synchronized void addManagedAdapter(ObjectAdapter adapter) {
      _managedAdapters.add(adapter);
    }
    
    synchronized void removeManagedAdapter(ObjectAdapter adapter) {
      _managedAdapters.remove(adapter);
    }
  }
    
  // the two classes for doing the hash table lookup.
  // note: given progressive keys p and q, and adapter values k and m
  // then p == q & p == k & q == m does not imply k == m,
  // the prefixes must match for equating adapter values, whereas the prefixes
  // are ignored in all other comparisons.
  
  static private class ProgressiveAIDKey {
    public ProgressiveAIDKey(byte [] aid, int [] offsets) {
      _aid = aid;
      _offsets = offsets;
      _hashes = new int[_offsets.length];
      
      // generate the hashes.
      _hash = 0;
      for(int i = 0; i < _hashes.length; ++i) {
        int s = _offsets[i];
        int e = ((i+1)<_hashes.length)?_offsets[i+1]:_aid.length;
        for(int j = s; j < e; ++j)
          _hash = 31*_hash + (_aid[j]&0xFF);
        _hashes[i] = _hash;
      }
      _hashTo = _offsets.length-1;
    }
    
    byte [] _aid;
    int [] _offsets;
    int [] _hashes;
    int _hash;
    int _hashTo;
    
    public boolean is_cacheable_object_key() {
      return _offsets[0] != 0;
    }
    
    public int hashCode() {
      return _hash;
    }
    
    public int useHash(int to) {
      _hashTo = to;
      _hash = (to==-1)?0:_hashes[to];
      return _hash;
    }
    
    public boolean equals(Object obj) {
      if(obj == this)
        return true;
      
      if(obj instanceof AdapterValue) {
        AdapterValue o2 = (AdapterValue)obj;
        
        // a whole key will not match a cacheable key.
        if((o2._offset == 0) != (_offsets[0] == 0))
          return false;
        
        // compare the hash codes
        if(o2._hash != _hash)
          return false;
        
        // compare the object keys
        int i = (_hashTo == _offsets.length-1)?_aid.length:_offsets[_hashTo+1];
        if(i-_offsets[0] != o2._aid.length-o2._offset)
          return false;
        int j = o2._aid.length;
        for(--i, --j; j >= o2._offset; --i, --j)
          if(_aid[i] != o2._aid[j])
            return false;
        return true;
      }
      
      // this is not a usual case
      if(obj instanceof ProgressiveAIDKey) {
        ProgressiveAIDKey o2 = (ProgressiveAIDKey)obj;
        
        // a whole key will not match a cacheable key.
        if((o2._offsets[0] == 0) != (_offsets[0] == 0))
          return false;
        
        // compare the hash codes
        if(o2._hashTo != this._hashTo)
          return false;
        for(int i = _hashTo; i >= 0; --i)
          if(o2._hashes[i] != _hashes[i] ||
          o2._offsets[i]-o2._offsets[0] != _offsets[i]-_offsets[0])
            return false;
        
        // compare the object keys
        int i = (_hashTo == _offsets.length-1)?_aid.length:_offsets[_hashTo+1];
        int j = (o2._hashTo == o2._offsets.length-1)?o2._aid.length:o2._offsets[o2._hashTo+1];
        for(--i, --j; i >= _offsets[0]; --i, --j)
          if(_aid[i] != o2._aid[j])
            return false;
        return true;
      }
      
      return false;
    }
  }
  
  private static class AdapterValue {
    public AdapterValue(byte [] aid,  int offset, ObjectAdapter adapter) {
      _aid = aid;
      _offset = offset;
      
      _hash = 0;
      for (int i = offset; i < _aid.length; ++i)
        _hash = 31*_hash + (_aid[i]&0xFF);
      
      _adapter = adapter;
      if(_adapter != null)
        _priority = _adapter.cache_priority();
      else
        _priority = -1;
    }
    
    byte [] _aid;
    int _hash;
    int _offset;
    ObjectAdapter _adapter;
    
    int _priority;
    int _hits = 0;
    
    MergeStack _destrRequests;
    
    boolean waiting_for_destroy = false;
    
    public boolean is_cacheable_object_key() {
      return _offset != 0;
    }
    
    public int hashCode() {
      return _hash;
    }
    
    public boolean equals(Object obj) {
      if(obj instanceof AdapterValue) {
        AdapterValue o2 = (AdapterValue)obj;
        
        if(o2._hash != _hash || o2._offset != _offset || o2._aid.length != _aid.length)
          return false;
        
        for(int i = _aid.length-1; i >= 0; --i)
          if(_aid[i] != o2._aid[i])
            return false;
        
        return true;
      }
      
      if(obj instanceof ProgressiveAIDKey) {
        // compare the hash codes
        ProgressiveAIDKey o2 = (ProgressiveAIDKey)obj;
        if(o2._hash != _hash)
          return false;
        
        // a whole key will not match a progressive key.
        if((_offset == 0) != (o2._offsets[0] == 0))
          return false;
        
        // compare the object keys
        int i = (o2._hashTo == o2._offsets.length-1)?o2._aid.length:o2._offsets[o2._hashTo+1];
        if(i-o2._offsets[0] != _aid.length-_offset)
          return false;
        int j = _aid.length;
        for(--i, --j; j >= _offset; --i, --j)
          if(o2._aid[i] != _aid[j])
            return false;
        return true;
      }

      return false;
    }
  }
}
